@@ -39,26 +39,64 @@ import rolenames
 CACHE_VALUES = True
 
 class Event:
-   """Dummy class for converting the source of an event to an
-   Accessible object.  We need this since the event object we
-   get from the atspi is read-only.  So, we create this dummy event
-   object to contain a copy of all the event members with the source
-   converted to an Accessible.  It is perfectly OK for event handlers
-   to annotate this object with their own attributes.
-   """
-   def __init__(self, e=None):
-       if e:
-           self.source   = Accessible.makeAccessible(e.source)
-           self.type     = e.type
-           self.detail1  = e.detail1
-           self.detail2  = e.detail2
-           self.any_data = e.any_data
-       else:
-           self.source   = None
-           self.type     = None
-           self.detail1  = None
-           self.detail2  = None
-           self.any_data = None
+    """Converts the source of an event to an Accessible object.  We
+    need this since the event object we get from the atspi is
+    read-only.  So, we create this dummy event object to contain a copy
+    of all the event members with the source converted to an
+    Accessible.  It is perfectly OK for event handlers to annotate this
+    object with their own attributes.
+    """
+
+    def getAnyData(e):
+        """Returns the any_data field, compensating for any
+        differences between pre-1.7.0 and post-1.7.0 implementations
+        of the AT-SPI."""
+        
+        # [[[TODO: WDW - HACK because AT-SPI 1.7.0 has
+        # introduced a binary incompatibility where the
+        # "any_data" of the event has changed from being a
+        # real "any_data" to an EventDetails structure that
+        # holds "any_data."  This check here helps us deal
+        # with the case where we might be talking to a
+        # pre-1.7.0 implementation or a 1.7.0+ implementation.
+        # What we're doing is just checking the CORBA typecode
+        # (as a string) to see if the type of the "any_data"
+        # field is this new structure.  If it is, we pull the
+        # "any_data" field from it.  If it isn't, we get the
+        # "any_data" field as we normally would.]]]
+        #
+        if e.any_data and (e.any_data.typecode().name) == "EventDetails":
+            return e.any_data.value().any_data
+        else:
+            return e.any_data
+       
+    getAnyData = staticmethod(getAnyData)
+
+    def __init__(self, e=None):
+        if e:
+            self.source   = Accessible.makeAccessible(e.source)
+            self.type     = e.type
+            self.detail1  = e.detail1
+            self.detail2  = e.detail2
+            
+            # If were talking to AT-SPI 1.7.0 or greater, we can get the
+            # application information right away because it is tucked in
+            # the EventDetails data new for 1.7.0.
+            #
+            if e.any_data and (e.any_data.typecode().name) == "EventDetails":
+                details = e.any_data.value()
+                self.any_data = details.any_data
+                if details.host_application:
+                    self.source.app = Accessible.makeAccessible(
+                        details.host_application)
+            else:
+                self.any_data = e.any_data
+        else:
+            self.source   = None
+            self.type     = None
+            self.detail1  = None
+            self.detail2  = None
+            self.any_data = None
 
 class Registry:
     """Delegates to the actual AT-SPI Regisitry.
@@ -105,15 +143,29 @@ class Registry:
         bonobo.main_quit()
 
     def registerEventListener(self, callback, eventType):
-	listener = EventListener(self.registry, callback, eventType)
+        """Registers the given eventType and callback with the Registry.
+
+        Arguments:
+        - callback: function to call with an AT-SPI event instance
+        - eventType: string representing the type of event
+        """
+       	listener = EventListener(self.registry, callback, eventType)
         self.__listeners.append(listener)
 
     def deregisterEventListener(self, callback, eventType):
+        """Unregisters the given eventType and callback with the Registry.
+
+        Arguments:
+        - callback: function to call with an AT-SPI event instance
+        - eventType: string representing the type of event
+        """
 	found = True
 	while len(self.__listeners) and found:
 	    for i in range(0, len(self.__listeners)):
 	        if (self.__listeners[i].callback == callback) \
 	           and (self.__listeners[i].eventType == eventType):
+                    # The __del__ method of the listener will unregister it.
+                    #
 	            self.__listeners.pop(i)
 		    found = True
 		    break
@@ -144,8 +196,8 @@ class Registry:
 class EventListener(Accessibility__POA.EventListener):
     """Registers a callback directly with the AT-SPI Registry for the
     given event type.  Most users of this module will not use this
-    class directly, but will instead use the addEventListener method.
-    """
+    class directly, but will instead use the registerEventListener method
+    of the Registry."""
 
     def __init__(self, registry, callback, eventType):
         self.registry  = registry
@@ -187,7 +239,7 @@ class KeystrokeListener(Accessibility__POA.DeviceEventListener):
     """Registers a callback directly with the AT-SPI Registry for the
     given keystroke.  Most users of this module will not use this
     class directly, but will instead use the registerKeystrokeListeners
-    method."""
+    method of the Registry."""
 
     def keyEventToString(event):
         return ("KEYEVENT: type=%d\n" % event.type) \
@@ -285,8 +337,12 @@ class Accessible:
     _cache = {}
 
     def init(registry):
-        # Register our various listeners.
-        #
+        """Registers various event listeners with the Registry to keep
+        the Accessible cache up to date.
+
+        Arguments:
+        - registry: an instance of Registry
+        """
         registry.registerEventListener(
             Accessible._onNameChanged,
             "object:property-change:accessible-name")
@@ -306,6 +362,12 @@ class Accessible:
     init = staticmethod(init)
 
     def shutdown(registry):
+        """Unregisters the event listeners that were registered in the
+        init method.
+
+        Arguments:
+        - registry: an instance of Registry
+        """
         registry.deregisterEventListener(
             Accessible._onNameChanged,
             "object:property-change:accessible-name")
@@ -335,12 +397,9 @@ class Accessible:
         if Accessible._cache.has_key(e.source):
             obj = Accessible._cache[e.source]
             if CACHE_VALUES:
-                obj.name = e.any_data.value()
-            try:
+                obj.name = Event.getAnyData(e).value()
+            if obj.__dict__.has_key("label"):
                 del obj.label
-            except:
-                debug.printException(debug.LEVEL_FINEST)
-                pass
 
     _onNameChanged = staticmethod(_onNameChanged)
 
@@ -355,12 +414,9 @@ class Accessible:
         if Accessible._cache.has_key(e.source):
             obj = Accessible._cache[e.source]
             if CACHE_VALUES:
-                obj.description = e.any_data.value()
-            try:
+                obj.description = Event.getAnyData(e).value()
+            if obj.__dict__.has_key("label"):
                 del obj.label
-            except:
-                debug.printException(debug.LEVEL_FINEST)
-                pass
 
     _onDescriptionChanged = staticmethod(_onDescriptionChanged)
 
@@ -434,6 +490,16 @@ class Accessible:
         if not acc:
             return obj
 
+        if isinstance(acc, Accessible):
+            debug.println(
+                debug.LEVEL_WARNING,
+                "WARNING: atspi.Accessible.makeAccessible:\n"
+                "         Parameter acc passed in is a\n" \
+                "         Python Accessible instead of an\n" \
+                "         AT-SPI Accessible.\n"
+                "         Returning Python Accessible.")
+            return acc
+            
 	# Comment these two lines out to eliminate any local caching 
 	# of Accessible objects.
 	#
@@ -671,8 +737,12 @@ class Accessible:
         """
 
         childCount = self._acc.childCount
-        if CACHE_VALUES:
-            self.childCount = childCount
+
+        # We don't want to cache this value because it's possible that it
+        # will continually change.
+        # if CACHE_VALUES:
+        #     self.childCount = childCount
+
         return childCount
 
     def __get_index(self):
@@ -680,8 +750,12 @@ class Accessible:
         """
 
         index = self._acc.getIndexInParent()
-        if CACHE_VALUES:
-            self.index = index
+
+        # We don't want to cache this value because it's possible that it
+        # will continually change.
+        # if CACHE_VALUES:
+        #     self.index = index
+
         return index
 
     def __get_role(self):
@@ -755,7 +829,7 @@ class Accessible:
                       "Finding app for source.name=" \
 	              + self.accessibleNameToString())
         obj = self
-        while obj.parent:
+        while obj.parent and (obj != obj.parent):
             obj = obj.parent
             debug.println(debug.LEVEL_FINEST,
                           "--> parent.name=" + obj.accessibleNameToString())
@@ -1091,16 +1165,16 @@ class Accessible:
 
         if includeApp:
 	    if self.app:
-                string = indent + " app.name=%-20s " \
+                string = indent + "app.name=%-20s " \
                          % self.app.accessibleNameToString()
 	    else:
-                string = indent + " app=None"
+                string = indent + "app=None "
         else:
             string = indent
 
         string += "name=%s role='%s' state='%s'" \
                   % (self.accessibleNameToString(),
-                     rolenames.getRoleName(self),
+                     self.role,
                      self.getStateString())
 
         return string
@@ -1412,6 +1486,75 @@ def getAcceleratorAndShortcut(obj):
 
     return [accelerator, fullShortcut]
 
+
+########################################################################
+#                                                                      #
+# Debug utilities                                                      #
+#                                                                      #
+########################################################################
+
+def printAncestry(child):
+   """Prints a hierarchical view of a child's ancestry."""
+   
+   if not child:
+       return
+   
+   objects = [child]
+   parent = child.parent
+   while parent and (parent.parent != parent):
+      objects.insert(0, parent)
+      parent = parent.parent
+
+   indent = ""
+   for object in objects:
+      print object.toString(indent + "+-", False)
+      indent += "  "
+
+def printHierarchy(root, ooi, indent="", onlyShowing=True, omitManaged=True):
+    """Prints the accessible hierarchy of all children
+
+    Arguments:
+    -indent:      Indentation string
+    -root:        Accessible where to start
+    -ooi:         Accessible object of interest
+    -onlyShowing: If True, only show children painted on the screen
+    -omitManaged: If True, omit children that are managed descendants
+    """
+
+    if not root:
+        return
+
+    if root == ooi:
+       print root.toString(indent + "(*)", False)
+    else:
+       print root.toString(indent + "+-", False)
+       
+    rootManagesDescendants = root.state.count(\
+        Accessibility.STATE_MANAGES_DESCENDANTS)
+    
+    for i in range(0, root.childCount):
+        child = root.child(i)
+        if child == root:
+            print indent + "  " + "WARNING CHILD == PARENT!!!"
+        elif not child:
+            print indent + "  " + "WARNING CHILD IS NONE!!!"
+        elif child.parent != root:
+            print indent + "  " + "WARNING CHILD.PARENT != PARENT!!!"
+        else:
+            paint = (not onlyShowing) \
+                    or (onlyShowing \
+                        and child.state.count(Accessibility.STATE_SHOWING))
+            paint = paint \
+                    and ((not omitManaged) \
+                         or (omitManaged and not rootManagesDescendants))
+
+            if paint:
+               printHierarchy(child,
+                              ooi,
+                              indent + "  ",
+                              onlyShowing,
+                              omitManaged)
+      
 ########################################################################
 #                                                                      #
 # Testing functions.                                                   #
@@ -1422,7 +1565,7 @@ def __printTopObject(child):
     parent = child
     while parent:
 	if not parent.parent:
-            print "RAW TOP:", parent.name, parent.getRoleName()
+            print "RAW TOP:", parent.name, parent.role
         parent = parent.parent
     if (child.parent):
         accessible = Accessible.makeAccessible(child)
@@ -1439,7 +1582,7 @@ def __printDesktops():
         for j in range(0, desktop.childCount):
             app = desktop.getChildAtIndex(j)
             print "    App %d: name=%s role=%s" \
-		  % (j, app.name, app.getRoleName())
+		  % (j, app.name, app.role)
 
 def __notifyEvent(event):
         print event.type, event.source.name, \
@@ -1447,7 +1590,7 @@ def __notifyEvent(event):
               event.any_data
 	__printTopObject(event.source)
 	if not event.source.parent:
-	    print "NO PARENT:", event.source.name, event.source.getRoleName()
+	    print "NO PARENT:", event.source.name, event.source.role
 
 def __notifyKeystroke(event):
     print "keystroke type=%d hw_code=%d modifiers=%d event_string=(%s) " \
