@@ -38,6 +38,7 @@ import script
 import settings
 import speech
 import speechgenerator
+import speechserver
 
 from orca_i18n import _                          # for gettext support
 
@@ -59,6 +60,7 @@ class Script(script.Script):
         script.Script.__init__(self, app)
 
         self.flatReviewContext = None
+        self.readTableCellRow = settings.getSetting(settings.READ_TABLE_CELL_ROW, False)
 
         ################################################################
         #                                                              #
@@ -92,14 +94,14 @@ class Script(script.Script):
                     Script.rightClickReviewItem,
                     _("Performs right click on current flat review item."))))
 
-        #self.keybindings.add(
-        #    keybindings.KeyBinding(
-        #        "KP_Add", \
-        #        0, \
-        #        0, \
-        #        input_event.InputEventHandler(\
-        #            Script.sayAll,
-        #            _("Speaks entire document."))))
+        self.keybindings.add(
+            keybindings.KeyBinding(
+                "KP_Add", \
+                0, \
+                0, \
+                input_event.InputEventHandler(\
+                    Script.sayAll,
+                    _("Speaks entire document."))))
 
         self.keybindings.add(
             keybindings.KeyBinding(
@@ -283,6 +285,15 @@ class Script(script.Script):
                 0, \
                 reviewNextCharacterHandler))
 
+        self.keybindings.add(
+            keybindings.KeyBinding(
+                "F11", \
+                1 << orca.MODIFIER_ORCA, \
+                1 << orca.MODIFIER_ORCA, \
+                input_event.InputEventHandler(\
+                    Script.toggleTableCellReadMode,
+                    _("Toggles whether to read just the current table cell or the whole row."))))
+
         ################################################################
         #                                                              #
         # Braille bindings                                             #
@@ -336,6 +347,9 @@ class Script(script.Script):
         self.listeners["object:property-change:accessible-name"] = \
             self.onNameChanged
 
+        self.listeners["mouse:button"]                           = \
+            self.onMouseButton
+
         self.listeners["object:text-caret-moved"]                = \
             self.onCaretMoved
         self.listeners["object:text-changed:delete"]             = \
@@ -433,6 +447,38 @@ class Script(script.Script):
 
         return speechgenerator.SpeechGenerator()
 
+    def __sayAllProgressCallback(self, context, type):
+        # [[[TODO: WDW - this needs work.  Need to be able to manage
+        # the monitoring of progress and couple that with both updating
+        # the visual progress of what is being spoken as well as
+        # positioning the cursor when speech has stopped.]]]
+        #
+        if type == speechserver.SayAllContext.PROGRESS:
+            #print "PROGRESS", context.utterance, context.currentOffset
+            #obj = context.obj
+            #[x, y, width, height] = obj.text.getCharacterExtents(
+            #    context.currentOffset, 0)
+            #print context.currentOffset, x, y, width, height
+            #orca.drawOutline(x, y, width, height)
+            pass
+        elif type == speechserver.SayAllContext.INTERRUPTED:
+            #print "INTERRUPTED", context.utterance, context.currentOffset
+            context.obj.text.setCaretOffset(context.currentOffset);
+        elif type == speechserver.SayAllContext.COMPLETED:
+            #print "COMPLETED", context.utterance, context.currentOffset
+            context.obj.text.setCaretOffset(context.currentOffset);
+    
+    def sayAll(self, inputEvent):
+        if not orca.locusOfFocus:
+            pass
+        elif orca.locusOfFocus.text:
+            speech.sayAll(textLines(orca.locusOfFocus),
+                          self.__sayAllProgressCallback)
+        else:
+            speech.speakUtterances(
+                self.speechGenerator.getSpeech(orca.locusOfFocus, False))
+        return True
+    
     def whereAmI(self, inputEvent):
         self.updateBraille(orca.locusOfFocus)
 
@@ -810,7 +856,8 @@ class Script(script.Script):
         # are on the very first line.  Otherwise, we show only the
         # line.
         #
-        if obj.role == rolenames.ROLE_TEXT:
+        if (obj.role == rolenames.ROLE_TEXT) \
+           or (obj.role == rolenames.ROLE_PARAGRAPH):
             text = obj.text
             [string, startOffset, endOffset] = text.getTextAtOffset(
                 text.caretOffset,
@@ -907,27 +954,15 @@ class Script(script.Script):
 
         orca.visualAppearanceChanged(event, event.source)
 
-    def onCaretMoved(self, event):
-        """Called whenever the caret moves.
+    def onMouseButton(self, event):
+        """Called whenever a mouse button is pressed or released.
 
         Arguments:
         - event: the Event
         """
-
-        # Ignore text deletions from non-focused objects, unless the
-        # currently focused object is the parent of the object from which
-        # text was deleted
-        #
-        if (event.source != orca.locusOfFocus) \
-               and (event.source.parent != orca.locusOfFocus):
-            return
-
-        # We always automatically go back to focus tracking mode when
-        # the caret moves in the focused object.
-        #
-        if self.flatReviewContext:
-            self.toggleFlatReviewMode()
-
+        orca.lastInputEvent = input_event.MouseButtonEvent(event)
+        
+    def _presentTextAtNewCaretPosition(self, event):
         # Magnify the object.  [[[TODO: WDW - this is a hack for now.]]]
         #
         #mag.magnifyAccessible(event.source)
@@ -948,17 +983,23 @@ class Script(script.Script):
         if brailleNeedsRepainting:
             self.updateBraille(event.source)
 
-        if (not orca.lastInputEvent) \
-            or \
-            (not isinstance(orca.lastInputEvent, input_event.KeyboardEvent)):
+        if not orca.lastInputEvent:
             return
-
+        
+        if isinstance(orca.lastInputEvent, input_event.MouseButtonEvent):
+            if not orca.lastInputEvent.pressed:
+                sayLine(event.source)
+            return
+            
         # Guess why the caret moved and say something appropriate.
         # [[[TODO: WDW - this motion assumes traditional GUI
         # navigation gestures.  In an editor such as vi, line up and
         # down is done via other actions such as "i" or "j".  We may
         # need to think about this a little harder.]]]
-        #
+        #            
+        if not isinstance(orca.lastInputEvent, input_event.KeyboardEvent):
+            return
+        
         string = orca.lastInputEvent.event_string
         mods = orca.lastInputEvent.modifiers
         controlMask = 1 << atspi.Accessibility.MODIFIER_CONTROL
@@ -982,7 +1023,30 @@ class Script(script.Script):
                 sayLine(event.source)
             else:
                 sayCharacter(event.source)
+        
+    def onCaretMoved(self, event):
+        """Called whenever the caret moves.
 
+        Arguments:
+        - event: the Event
+        """
+
+        # Ignore text deletions from non-focused objects, unless the
+        # currently focused object is the parent of the object from which
+        # text was deleted
+        #
+        if (event.source != orca.locusOfFocus) \
+               and (event.source.parent != orca.locusOfFocus):
+            return
+
+        # We always automatically go back to focus tracking mode when
+        # the caret moves in the focused object.
+        #
+        if self.flatReviewContext:
+            self.toggleFlatReviewMode()
+
+        self._presentTextAtNewCaretPosition(event)
+        
     def onTextDeleted(self, event):
         """Called whenever text is deleted from an object.
 
@@ -1052,8 +1116,23 @@ class Script(script.Script):
 
         child = atspi.Accessible.makeAccessible(event.any_data.value())
 
+        # [[[TODO: WDW - this is an odd hacky thing I've somewhat drawn
+        # from Gnopernicus.  The notion here is that we get an active
+        # descendant changed event, but that object tends to have children
+        # itself and we need to decide what to do.  Well...the idea here
+        # is that the last child (Gnopernicus chooses child(1)), tends to
+        # be the child with information.  The previous children tend to
+        # be non-text or just there for spacing or something.  You will
+        # see this in the various table demos of gtk-demo and you will
+        # also see this in the Contact Source Selector in Evolution.
+        #
+        # Just note that this is most likely not a really good solution
+        # for the general case.  That needs more thought.  But, this
+        # comment is here to remind us this is being done in poor taste
+        # and we need to eventually clean up our act.]]]
+        #
         if child.childCount:
-            orca.setLocusOfFocus(event, child.child(1))
+            orca.setLocusOfFocus(event, child.child(child.childCount - 1))
         else:
             orca.setLocusOfFocus(event, child)
 
@@ -1217,6 +1296,14 @@ class Script(script.Script):
 
         return flat_review.clusterZonesByLine(zones)
 
+    def toggleTableCellReadMode(self, inputEvent=None):
+        """Toggles an indicator for whether we should just read the current 
+        table cell or read the whole row."""
+
+        self.readTableCellRow = not self.readTableCellRow
+
+        return True
+
     def getFlatReviewContext(self):
         """Returns the flat review context, creating one if necessary."""
 
@@ -1404,7 +1491,8 @@ class Script(script.Script):
             self.targetCursorCell = 1
             self.updateBrailleReview(self.targetCursorCell)
         elif braille.beginningIsShowing and orca.locusOfFocus \
-             and (orca.locusOfFocus.role == rolenames.ROLE_TEXT):
+             and ((orca.locusOfFocus.role == rolenames.ROLE_TEXT) \
+                  or (orca.locusOfFocus.role == rolenames.ROLE_PARAGRAPH)):
             # If we're at the beginning of a line of a multiline text
             # area, then force it's caret to the end of the previous
             # line.  The assumption here is that we're currently
@@ -1463,7 +1551,8 @@ class Script(script.Script):
             self.targetCursorCell = 1
             self.updateBrailleReview(self.targetCursorCell)
         elif braille.endIsShowing and orca.locusOfFocus \
-             and (orca.locusOfFocus.role == rolenames.ROLE_TEXT):
+             and ((orca.locusOfFocus.role == rolenames.ROLE_TEXT) \
+                  or (orca.locusOfFocus.role == rolenames.ROLE_PARAGRAPH)):
             # If we're at the end of a line of a multiline text area, then
             # force it's caret to the beginning of the next line.  The
             # assumption here is that we're currently viewing the line that
@@ -1807,7 +1896,94 @@ class Script(script.Script):
 # Functions for handling output of AccessibleText objects to speech.   #
 #                                                                      #
 ########################################################################
+        
+def textLines(obj):
+    """Creates a generator that can be used to iterate over each line
+    of a text object, starting at the caret offset.  One can do something
+    like the following to access the lines:
 
+    for line in atspi.textLines(orca.locusOfFocus.text):
+        <<<do something with the line>>>
+
+    Arguments:
+    - obj: an Accessible that has a text specialization
+
+    Returns a tuple: [SayAllContext, acss], where SayAllContext has the
+    text to be spoken and acss is an ACSS instance for speaking the text.
+    """
+    if not obj:
+        return
+
+    text = obj.text
+    if not text:
+        return
+
+    length = text.characterCount
+    offset = text.caretOffset
+    
+    # Get the next line of text to read 	 
+    #
+    lastEndOffset = -1
+    while offset < length:
+        [string, startOffset, endOffset] = text.getTextAtOffset(
+            offset,
+            atspi.Accessibility.TEXT_BOUNDARY_LINE_START)
+
+        # [[[WDW - HACK: this is here because getTextAtOffset
+        # tends not to be implemented consistently across toolkits.
+        # Sometimes it behaves properly (i.e., giving us an endOffset
+        # that is the beginning of the next line), sometimes it
+        # doesn't (e.g., giving us an endOffset that is the end of
+        # the current line).  So...we hack.  The whole 'max' deal
+        # is to account for lines that might be a brazillion lines
+        # long.]]]
+        #
+        if endOffset == lastEndOffset:
+            offset = max(offset + 1, lastEndOffset + 1)
+            lastEndOffset = endOffset
+            continue
+
+        lastEndOffset = endOffset
+        offset = endOffset
+
+        # Strip trailing new lines
+        #
+        #if string[-1:] == "\n":
+        #    string = string[0][:-1]
+
+        yield [speechserver.SayAllContext(obj, string, startOffset, endOffset),
+               None]
+
+def getLinkIndex(obj, characterIndex):
+    """A brute force method to see if an offset is a link.  This
+    is provided because not all Accessible Hypertext implementations
+    properly support the getLinkIndex method.  Returns an index of
+    0 or greater of the characterIndex is on a hyperlink.
+
+    Arguments:
+    -obj: the Accessible object with the Accessible Hypertext specialization
+    -characterIndex: the text position to check
+    """
+    
+    if not obj:
+        return -1
+
+    text = obj.text
+    if not text:
+        return -1
+
+    hypertext = obj.hypertext
+    if not hypertext:
+        return -1
+
+    for i in range(0, hypertext.getNLinks()):
+        link = hypertext.getLink(i)
+        if (characterIndex >= link.startIndex) \
+           and (characterIndex <= link.endIndex):
+            return i
+
+    return -1
+    
 def sayLine(obj):
     """Speaks the line of an AccessibleText object that contains the
     caret. [[[TODO: WDW - what if the line is empty?]]]
@@ -1819,8 +1995,15 @@ def sayLine(obj):
 
     # Get the AccessibleText interface of the provided object
     #
-    result = atspi.getTextLineAtCaret(obj)
-    speech.speak(result[0])
+    [line, startOffset, endOffset] = atspi.getTextLineAtCaret(obj)
+
+    voices = settings.getSetting(settings.VOICES, None)
+    if line.isupper():
+        voice = voices[settings.UPPERCASE_VOICE]
+    else:
+        voice = voices[settings.DEFAULT_VOICE]
+
+    speech.speak(line, voice)
 
 def sayWord(obj):
     """Speaks the word at the caret.  [[[TODO: WDW - what if there is no
@@ -1833,9 +2016,20 @@ def sayWord(obj):
 
     text = obj.text
     offset = text.caretOffset
-    word = text.getTextAtOffset(offset,
-                                atspi.Accessibility.TEXT_BOUNDARY_WORD_START)
-    speech.speak(word[0])
+    [word, startOffset, endOffset] = \
+        text.getTextAtOffset(offset,
+                             atspi.Accessibility.TEXT_BOUNDARY_WORD_START)
+
+    voices = settings.getSetting(settings.VOICES, None)
+    if getLinkIndex(obj, offset) >= 0:
+        print "HYPERLINK!"
+        voice = voices[settings.HYPERLINK_VOICE]
+    elif word.isupper():
+        voice = voices[settings.UPPERCASE_VOICE]
+    else:
+        voice = voices[settings.DEFAULT_VOICE]
+
+    speech.speak(word, voice)
 
 def sayCharacter(obj):
     """Speak the character under the caret.  [[[TODO: WDW - isn't the
@@ -1849,11 +2043,17 @@ def sayCharacter(obj):
     text = obj.text
     offset = text.caretOffset
     character = text.getText(offset, offset+1)
-    if character.isupper():
-        voices = settings.getSetting(settings.VOICES, None)
-        speech.speak(character, voices[settings.UPPERCASE_VOICE])
+
+    voices = settings.getSetting(settings.VOICES, None)
+    if getLinkIndex(obj, offset) >= 0:
+        print "HYPERLINK!"
+        voice = voices[settings.HYPERLINK_VOICE]
+    elif character.isupper():
+        voice = voices[settings.UPPERCASE_VOICE]
     else:
-        speech.speak(character)
+        voice = voices[settings.DEFAULT_VOICE]
+
+    speech.speak(character, voice)
 
 # Dictionary that defines the state changes we care about for various
 # objects.  The key represents the role and the value represents a list
@@ -1861,6 +2061,8 @@ def sayCharacter(obj):
 #
 state_change_notifiers = {}
 
+state_change_notifiers[rolenames.ROLE_CHECK_MENU_ITEM] = ("checked",
+                                                        None)
 state_change_notifiers[rolenames.ROLE_CHECK_BOX]     = ("checked",
                                                         None)
 state_change_notifiers[rolenames.ROLE_PANEL]         = ("showing",
