@@ -142,27 +142,36 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             # a syntax error in it, so we want to give script writers
             # a vehicle for debugging these types of things.
             #
-            scriptPackages = settings.getSetting(settings.SCRIPT_PACKAGES,
-                                                 ["orca-scripts", "scripts"])
+            scriptPackages = settings.getSetting(
+                settings.SCRIPT_PACKAGES,
+                ["", "orca-scripts", "scripts"])
 
             moduleName = settings.getScriptModuleName(app)
             module = None
-            
-            for package in scriptPackages:
-                if len(package):
-                    name = package + "." + moduleName
-                else:
-                    name = moduleName
-
-                try:
-                    module = __import__(name,
-                                        globals(),
-                                        locals(),
-                                        [''])
-                    break
-                except:
-                    debug.printException(debug.LEVEL_FINEST)
 
+            if moduleName and len(moduleName):
+                for package in scriptPackages:
+                    if len(package):
+                        name = package + "." + moduleName
+                    else:
+                        name = moduleName
+                    try:
+                        debug.println(debug.LEVEL_FINEST,
+                                      "Looking for script at %s.py..." % name)
+                        module = __import__(name,
+                                            globals(),
+                                            locals(),
+                                            [''])
+                        debug.println(debug.LEVEL_FINEST,
+                                      "...found %s.py" % name)
+                        break
+                    except ImportError:
+                        debug.println(debug.LEVEL_FINEST,
+                                      "...could not find %s.py" % name)
+                    except:
+                        debug.printException(debug.LEVEL_SEVERE)
+                        debug.println(debug.LEVEL_SEVERE,
+                                      "While attempting to import %s" % name)
             if module:
                 try:
                     script = module.Script(app)
