@@ -95,6 +95,8 @@ class SpeechGenerator:
              self._getSpeechForPageTab
         self.speechGenerators[rolenames.ROLE_PAGE_TAB_LIST]       = \
              self._getSpeechForPageTabList
+        self.speechGenerators[rolenames.ROLE_PARAGRAPH]           = \
+             self._getSpeechForText
         self.speechGenerators[rolenames.ROLE_PASSWORD_TEXT]       = \
              self._getSpeechForText
         self.speechGenerators[rolenames.ROLE_PROGRESS_BAR]        = \
@@ -203,7 +205,8 @@ class SpeechGenerator:
         text = obj.label
 
         if verbosity == settings.VERBOSITY_LEVEL_VERBOSE:
-            text += " " + getSpeechForRoleName(obj)
+            if (obj.label == None) or (obj.role != rolenames.ROLE_UNKNOWN):
+                text += " " + getSpeechForRoleName(obj)
 
         return [text]
 
@@ -381,13 +384,13 @@ class SpeechGenerator:
         Returns a list of utterances to be spoken for the object.
         """
 
-        utterances = self._getSpeechForCheckBox(obj, False)
+        utterances = self._getSpeechForCheckBox(obj, already_focused)
 
         verbosity = settings.getSetting(settings.SPEECH_VERBOSITY_LEVEL,
                                         settings.VERBOSITY_LEVEL_VERBOSE)
 
-        if verbosity == settings.VERBOSITY_LEVEL_VERBOSE:
-            utterances.extend(self._getSpeechForAvailability(obj))
+        if (verbosity == settings.VERBOSITY_LEVEL_VERBOSE) \
+           and not already_focused:
             utterances.extend(self._getSpeechForAccelerator(obj))
 
         self._debugGenerator("_getSpeechForCheckMenuItem",
@@ -443,6 +446,9 @@ class SpeechGenerator:
         #
         # Preference is given to #1, if it exists.
         #
+        # If the label of the combo box is the same as the utterance for
+        # the child object, then this utterance is only spoken once.
+        #
         # [[[TODO: WDW - Combo boxes are complex beasts.  This algorithm
         # needs serious work.  Logged as bugzilla bug 319745.]]]
         #
@@ -458,7 +464,8 @@ class SpeechGenerator:
         if textObj:
             result = atspi.getTextLineAtCaret(textObj)
             line = result[0]
-            utterances.append(line)
+            if line != obj.label:
+                utterances.append(line)
         else:
             selectedItem = None
             comboSelection = obj.selection
@@ -466,12 +473,14 @@ class SpeechGenerator:
                 selectedItem = atspi.Accessible.makeAccessible(\
                     comboSelection.getSelectedChild(0))
             if selectedItem:
-                utterances.append(selectedItem.label)
+                if selectedItem.label != obj.label:
+                    utterances.append(selectedItem.label)
             else:
                 result = atspi.getTextLineAtCaret(obj)
                 selectedText = result[0]
                 if len(selectedText) > 0:
-                    utterances.append(selectedText)
+                    if selectedText != obj.label:
+                        utterances.append(selectedText)
                 else:
                     debug.println(
                         debug.LEVEL_SEVERE,
@@ -1463,7 +1472,9 @@ class SpeechGenerator:
             if parent == stopAncestor:
                 break
             if (parent.role != rolenames.ROLE_FILLER) \
-                and (parent.role != rolenames.ROLE_SPLIT_PANE):
+                and (parent.role != rolenames.ROLE_SPLIT_PANE) \
+                and (parent.role != rolenames.ROLE_SCROLL_PANE) \
+                and (parent.role != rolenames.ROLE_UNKNOWN):
                 if len(parent.label) > 0:
                     utterances.append(parent.label + " " \
                                       + getSpeechForRoleName(parent))
